using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace MyDirect3DProject
{
	/// <summary>
	/// This is the main class of my Direct3D application
	/// </summary>
	public class MainClass : Form
	{
		/// <summary>
		/// The rendering device
		/// </summary>
		private Device device;
		private VertexBuffer vb;
		private IndexBuffer ib;
		private const int Rows = 42;
		private const int Columns = 43;

		private long frames;
		private System.DateTime startTime;
		
		private Effect myEffect;
		private VertexElement[] myVertexElement;
		
		public MainClass()
		{
			this.ClientSize = new System.Drawing.Size(640, 480);
			this.Text = "Direct3D Project";
			this.SetStyle(ControlStyles.AllPaintingInWmPaint|ControlStyles.Opaque, true);
		}
		
		public bool InitializeGraphics()
		{
			try {
				startTime = System.DateTime.Now;
				
				// Now let's setup the Direct3D stuff
				PresentParameters presentParams = new PresentParameters();
				presentParams.Windowed = true;
				presentParams.SwapEffect = SwapEffect.Discard;
				presentParams.EnableAutoDepthStencil = true;			
				presentParams.AutoDepthStencilFormat = DepthFormat.D16;
				
				// Create the device
				device = new Device(0, DeviceType.Hardware, this,
				                    CreateFlags.HardwareVertexProcessing, presentParams);
				
				// Setup the event handlers for the device
				device.DeviceLost     += new EventHandler(this.InvalidateDeviceObjects);
				device.DeviceReset    += new EventHandler(this.RestoreDeviceObjects);
				device.Disposing      += new EventHandler(this.DeleteDeviceObjects);
				device.DeviceResizing += new CancelEventHandler(this.EnvironmentResizing);
				
				vb = new VertexBuffer(typeof(CustomVertex.PositionNormal),
				                                   (Rows+1)*(Columns+1), device, Usage.WriteOnly,
				                                   0, Pool.Managed);
				CustomVertex.PositionNormal[] vtx = (CustomVertex.PositionNormal[]) vb.Lock(0, LockFlags.None);
				for(int i = 0; i < (Rows+1)*(Columns+1); i++)
				{
					float x = 2.0f*(i % (Columns+1))/(float)Columns - 1.0f;
					float y = 2.0f*(i / (Columns+1))/(float)Rows - 1.0f;
					float z =  0.5f+0.2f*(float)Math.Sin(8.0*x);
					float nx = 0.2f*8.0f*(float)Math.Cos(8.0*x);
					float ny = 0.0f;
					float nz = -1.0f;
					vtx[i] = new CustomVertex.PositionNormal(x, y, z, nx, ny, nz);
				}
				vb.Unlock();			
				
				ib = new IndexBuffer(typeof(short), 2*3*Rows*Columns, device, Usage.WriteOnly, Pool.Managed);
				short[] ind = (short[]) ib.Lock(0, LockFlags.None);
				for(int i = 0; i < Rows*Columns; i++) // zählt die Vierecke (= je zwei Dreiecke) durch
				{
					// Es gibt pro Zeile einen Vertex mehr (Columns + 1) als Vierecke!
					int iBase = (i % Columns) + (i / Columns)*(Columns + 1);
					ind[6*i  ] = (short)(iBase);
					ind[6*i+1] = (short)(iBase + Columns + 1);
					ind[6*i+2] = (short)(iBase + Columns + 2);
					ind[6*i+3] = (short)(iBase + Columns + 2);
					ind[6*i+4] = (short)(iBase + 1);
					ind[6*i+5] = (short)(iBase);					
				}
				ib.Unlock();

				// zwei Verzeichnisse hoch, damit es auch aus der Entwicklungsumgebung läuft
				myEffect = Effect.FromFile(device, @"..\..\phong.fx", null, ShaderFlags.None, null);
				myEffect.Technique = "dx9";
							
				myVertexElement = new VertexElement[]
				{
					new VertexElement(0, 0, DeclarationType.Float3, DeclarationMethod.Default, DeclarationUsage.Position, 0),
					new VertexElement(0, 3*4, DeclarationType.Float3, DeclarationMethod.Default, DeclarationUsage.Normal, 0),
					VertexElement.VertexDeclarationEnd
				};
			
				return true;
			} catch (DirectXException) {
				return false;
			}			
		}
		
		protected virtual void InvalidateDeviceObjects(object sender, EventArgs e)
		{		
		}
		
		protected virtual void RestoreDeviceObjects(object sender, EventArgs e)
		{		
		}
		
		protected virtual void DeleteDeviceObjects(object sender, EventArgs e)
		{
			// passiert hier auch alles automatisch;
			// manchmal ist es aber sinnvoller oder sogar nötig,
			// Objekte vor Programmende zu entsorgen.
			// Die Garbage Collection wird ggf. daran gehindert,
			// weil das device noch Referenzen auf diese Objekte hat.
			myEffect.Dispose();
			ib.Dispose();
			vb.Dispose();	
		}
		
		protected virtual void EnvironmentResizing(object sender, CancelEventArgs e)
		{
		}
		
		/// <summary>
		/// This method moves the scene
		/// </summary>
		protected virtual void FrameMove()
		{
			// TODO : Frame movement
		}
		
		/// <summary>
		/// This method renders the scene
		/// </summary>
		protected virtual void Render()
		{
			if (device != null) {

				double time = (System.DateTime.Now - startTime).TotalSeconds;
				
				device.Clear(ClearFlags.Target|ClearFlags.ZBuffer, Color.Blue, 1.0f, 0);
				device.BeginScene();
				
				device.SetStreamSource(0, vb, 0);
				device.Indices = ib;
				
				// "new" ist hier unschön, eigentlich nur nach Verlust des device neu aufbauen 
				device.VertexDeclaration = new VertexDeclaration(device, myVertexElement);				

				device.RenderState.NormalizeNormals = true; // schneller ist, mit Normalen der Länge 1 zu arbeiten					
				
				Matrix proj = Matrix.PerspectiveFovLH(0.8f, Width/(float)Height, 0.1f, 50.0f);
				Matrix view = Matrix.LookAtLH(new Vector3(0.0f, 0.0f, -2.0f), new Vector3(), new Vector3(0.0f, 1.0f, 0.0f));
				Matrix world = Matrix.RotationX(1.2f)*Matrix.RotationY(0.8f*(float)time);
				Matrix worldT = world;
				worldT.Transpose(worldT);
				Matrix worldIT = worldT;
				worldIT.Invert();
				Matrix viewT = view;
				viewT.Transpose(viewT);
				Matrix viewIT = viewT;
				viewIT.Invert();

				myEffect.SetValue("world", world);
				myEffect.SetValue("wvp", world*view*proj);
				myEffect.SetValue("worldIT", worldIT);
				myEffect.SetValue("viewIT", viewIT);
				myEffect.SetValue("lightPos", new Vector4(-5f, 7f, -4f, 1.0f));				
				myEffect.SetValue("diffColor", new Vector4(0.8f, 0.4f, 0.9f, 1.0f));
				myEffect.SetValue("specExpon", 30f);

				int numPasses = myEffect.Begin(0);
				for(int pass = 0; pass < numPasses; pass++)
				{
					myEffect.Pass(pass);
					device.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, (Rows+1)*(Columns+1), 0, 2*Rows*Columns);
				}
				myEffect.End();
				
				device.EndScene();
				device.Present();
				
				device.VertexDeclaration.Dispose();
				
				// unschön: irgendwann nach einer halben Ewigkeit läuft die Variable frames über
				frames ++;
				Text = (frames/time).ToString("###.0") + " fps";
			}
		}
		
		/// <summary>
		/// Our main loop
		/// </summary>
		public void Run()
		{
			// While the form is still valid, render and process messages
			while (Created) {
				FrameMove();	
				Render();
				Application.DoEvents();
			}
		}
		
		protected override void OnPaint(PaintEventArgs e)
		{
			this.Render();
		}
		
		protected override void OnKeyPress(KeyPressEventArgs e)
		{
			base.OnKeyPress(e);
			if ((int)e.KeyChar == (int)System.Windows.Forms.Keys.Escape) {
				this.Close(); 
			}
		}
		
		/// <summary>
		/// The main entry point for the application
		/// </summary>
		static void Main()
		{
			using (MainClass mainClass = new MainClass()) {
				if (!mainClass.InitializeGraphics()) {
					MessageBox.Show("Error while initializing Direct3D");
					return;
				}
				mainClass.Show();
				mainClass.Run();
			}
		}
	}
}
